use adw::{prelude::*, subclass::prelude::*};
use cairo::Context;
use glib::Object;
use kifu_core::ui::Node;
use std::{cell::RefCell, f64::consts::PI, rc::Rc};

const NODE_SIZE: f64 = 10.;
const NODE_ROW_HEIGHT: f64 = 30.;
const NODE_COLUMN_WIDTH: f64 = 30.;

#[derive(Default)]
pub struct ReviewTreePrivate {
    tree: Rc<RefCell<Option<Node>>>,
}

#[glib::object_subclass]
impl ObjectSubclass for ReviewTreePrivate {
    const NAME: &'static str = "ReviewTree";
    type Type = ReviewTree;
    type ParentType = gtk::DrawingArea;
}

impl ObjectImpl for ReviewTreePrivate {}
impl WidgetImpl for ReviewTreePrivate {}
impl DrawingAreaImpl for ReviewTreePrivate {}

glib::wrapper! {
    pub struct ReviewTree(ObjectSubclass<ReviewTreePrivate>)
        @extends gtk::DrawingArea, gtk::Widget;
}

impl ReviewTree {
    pub fn new() -> Self {
        let s: Self = Object::builder().build();

        s.set_draw_func({
            let s = s.clone();
            move |_, context, width, height| {
                println!("drawing area: {} {}", width, height);
                let style_context = WidgetExt::style_context(&s);
                let bg = style_context.lookup_color("view_bg_color").unwrap();
                let fg = style_context.lookup_color("view_fg_color").unwrap();

                context.set_source_rgb(bg.red() as f64, bg.green() as f64, bg.blue() as f64);
                let _ = context.paint();

                if let Some(tree) = &*s.imp().tree.borrow() {
                    let (width, height) = max_tree_dimensions(tree);
                    println!("tree dimensions: {} {}", width, height);
                    s.set_width_request(
                        width as i32 * (NODE_SIZE as i32 + NODE_COLUMN_WIDTH as i32) + 20,
                    );
                    s.set_height_request(
                        height as i32 * (NODE_SIZE as i32 + NODE_ROW_HEIGHT as i32) + 20,
                    );

                    context.set_source_rgb(fg.red() as f64, fg.green() as f64, fg.blue() as f64);
                    draw_tree(&context, &tree);
                }
            }
        });

        s
    }

    pub fn set_tree(&self, tree: Node) {
        *self.imp().tree.borrow_mut() = Some(tree);
        self.queue_draw();
    }
}

impl Default for ReviewTree {
    fn default() -> Self {
        Self::new()
    }
}

fn draw_node(context: &Context, x: f64, y: f64) {
    context.arc(x, y, NODE_SIZE, 0., 2. * PI);
    let _ = context.fill();
}

fn draw_tree(context: &Context, tree: &Node) {
    let mut row: Vec<&Node> = vec![];
    let mut next_row: Vec<&Node> = vec![];
    let mut width: Vec<usize> = vec![];

    let mut x = 0;
    let mut y = 0;

    row.push(tree);
    width.push(1);

    while row.len() != 0 {
        for node in row.into_iter() {
            draw_node(
                context,
                10. + (x as f64) * (NODE_SIZE + NODE_COLUMN_WIDTH),
                10. + (y as f64) * (NODE_SIZE + NODE_ROW_HEIGHT),
            );
            next_row.append(&mut node.children.iter().map(|n| n).collect::<Vec<&Node>>());
            x = x + 1;
        }
        x = 0;
        y = y + 1;
        row = next_row;
        next_row = vec![];
    }
}

fn max_tree_dimensions(tree: &Node) -> (usize, usize) {
    let mut row: Vec<&Node> = vec![];
    let mut next_row: Vec<&Node> = vec![];
    let mut width: Vec<usize> = vec![];

    row.push(tree);
    width.push(1);

    while row.len() != 0 {
        println!("new row");
        for node in row.into_iter() {
            println!(
                "{:?} {:?} {}",
                node.color,
                node.position,
                node.children.len()
            );
            next_row.append(&mut node.children.iter().map(|n| n).collect::<Vec<&Node>>());
        }
        width.push(next_row.len());
        row = next_row;
        next_row = vec![];
    }

    (
        width.iter().fold(0, |a, b| if a > *b { a } else { *b }),
        width.len(),
    )
}
